<?php

require_once DRUPAL_ROOT . '/includes/cache.inc';

class SchrodiCacheDatabaseCache extends DrupalDatabaseCache {
  /**
   * Implements DrupalCacheInterface::set().
   */
  function set($cid, $data, $expire = CACHE_PERMANENT) {
    $fields = array(
      'serialized' => 0,
      // This is the whole reason for this class - use better than second
      // granularity, because it makes the hit rate way better.
      'created' => round(microtime(TRUE), 3),
      'expire' => $expire,
    );
    if (!is_string($data)) {
      $fields['data'] = serialize($data);
      $fields['serialized'] = 1;
    }
    else {
      $fields['data'] = $data;
      $fields['serialized'] = 0;
    }
    try {
      db_merge($this->bin)
        ->key(array('cid' => $cid))
        ->fields($fields)
        ->execute();
    }
    catch (Exception $e) {
      // The database may not be available, so we'll ignore cache_set requests.
    }
  }
}

class SchrodiCacheApcCache implements DrupalCacheInterface {

  /**
   * The apc prefix.
   */
  protected $apcPrefix;

  /**
   * Constructs a DrupalDatabaseCache object.
   *
   * @param $bin
   *   The cache bin for which the object is created.
   */
  function __construct($bin) {
    global $conf;

    $this->apcPrefix = '';
    if (isset($conf['schrodicache']['apc_prefix'])) {
      $this->apcPrefix = $conf['schrodicache']['apc_prefix'];
    }
    $this->apcPrefix .= $bin;
  }

  /**
   * Implements DrupalCacheInterface::get().
   */
  function get($cid) {
    return apc_fetch($this->apcPrefix . $cid);
  }

  /**
   * Implements DrupalCacheInterface::set().
   */
  function set($cid, $data, $expire = CACHE_PERMANENT) {
    $fields = (object) array(
      'serialized' => 0,
      'created' => round(microtime(TRUE), 3),
      'expire' => $expire,
      'cid' => $cid,
      'data' => $data,
      'serialized' => 0,
    );
    apc_store($this->apcPrefix . $cid, $fields);
  }

  /**
   * Implements DrupalCacheInterface::isEmpty().
   */
  function isEmpty() {
    $apc = new ApcIterator('user', '/^' . preg_quote($this->apcPrefix) . '/');
    return $apc->getTotalCount > 0;
  }

  /**
   * Implements DrupalCacheInterface::clear().
   */
  function clear($cid = NULL, $wildcard = FALSE) {
    if (!empty($cid)) {
      if ($wildcard) {
        if ($cid == '*') {
          apc_delete(new APCIterator('user', '/^' . preg_quote($this->apcPrefix) . '/'));
        }
        else {
          apc_delete(new APCIterator('user', '/^' . preg_quote($this->apcPrefix . $cid) . '/'));
        }
      }
      else {
        apc_delete($cid);
      }
    }
    else {
      apc_delete(new APCIterator('user', '/^' . preg_quote($this->apcPrefix) . '/'));
    }
  }

  /**
   * Implements DrupalCacheInterface::getMultiple().
   */
  function getMultiple(&$cids) {
    $cache = array();
    $prefixed_cids = array();
    foreach ($cids as $cid) {
      $prefixed_cids[$this->apcPrefix . $cid] = $cid;
    }
    $result = apc_fetch(array_keys($prefixed_cids));
    foreach ($result as $item) {
      $cache[$item->cid] = $item;
    }
    $cids = array_diff($cids, array_keys($cache));
    return $cache;
  }
}

class SchrodiCacheCache implements DrupalCacheInterface {

  /**
   * Collector for all last write times across all bins using SchrodiCache.
   */
  public static $writeTimes = array();

  /**
   * Prefix for bin-specific entry to track the last write.
   */
  const LAST_WRITE_TIME_PREFIX = 'last_write_time_';

  /**
   * The last write time for this bin.
   */
  protected $lastWriteTime;

  /**
   * The consistent cache backend.
   */
  protected $consistentBackend;

  /**
   * The inconsistent cache backend.
   */
  protected $inconsistentBackend;

  /**
   * Whether to update last write times immediately, or once at request end.
   */
  protected $immediateLastWriteTimeUpdates;

  /**
   * Constructs a SchrodiCacheCache object.
   *
   * @param $bin
   *   The cache bin for which the object is created.
   */
  function __construct($bin) {
    global $conf;

    $this->bin = $bin;

    // Default to more correct, tunable to less correct but faster.
    $this->immediateLastWriteTimeUpdates = TRUE;
    if (isset($conf['schrodicache']['immediate_last_write_time_updates'])) {
      $this->immediateLastWriteTimeUpdates = $conf['schrodicache']['immediate_last_write_time_updates'];
    }

    $consistent_class = 'SchrodiCacheDatabaseCache';
    if (!empty($conf['schrodicache']['consistent_backend'])) {
      $consistent_class = $conf['schrodicache']['consistent_backend'];
    }
    $this->consistentBackend = new $consistent_class($bin);

    $inconsistent_class = 'SchrodiCacheApcCache';
    if (!empty($conf['schrodicache']['inconsistent_backend'])) {
      $inconsistent_class = $conf['schrodicache']['inconsistent_backend'];
    }
    $this->inconsistentBackend = new $inconsistent_class($bin);

    $shutdown_function = array('SchrodiCacheCache', 'shutdownUpdateLastWriteTimes');
    $not_found = TRUE;
    $callbacks = drupal_register_shutdown_function();
    foreach ($callbacks as $info) {
      if ($info['callback'] == $shutdown_function) {
        $not_found = FALSE;
        break;
      }
    }
    if ($not_found) {
      drupal_register_shutdown_function($shutdown_function);
    }
    reset($callbacks);
  }

  public static function shutdownUpdateLastWriteTimes() {
    if (!empty(static::$writeTimes)) {
      static::updateLastWriteTimes(static::$writeTimes);
    }
  }

  /**
   * Implements DrupalCacheInterface::isEmpty().
   */
  function isEmpty() {
    return $this->consistentBackend->isEmpty();
  }

  /**
   * Implements DrupalCacheInterface::get().
   */
  function get($cid) {
    if ($last_write_time = $this->lastWriteTime()) {
      $item = $this->inconsistentBackend->get($cid);
      if ($item && $item->created >= $last_write_time) {
        return $item;
      }
    }

    if ($item = $this->consistentBackend->get($cid)) {
      $this->inconsistentBackend->set($item->cid, $item->data);
    }
    return $item;
  }

  /**
   * Implements DrupalCacheInterface::get().
   */
  function set($cid, $data, $expire = CACHE_PERMANENT) {
    $this->markAsOutdated();
    $this->consistentBackend->set($cid, $data, $expire);
    $this->inconsistentBackend->set($cid, $data, $expire);
  }

  /**
   * Implements DrupalCacheInterface::clear().
   */
  function clear($cid = NULL, $wildcard = FALSE) {
    $this->consistentBackend->clear($cid, $wildcard);
    $this->inconsistentBackend->clear($cid, $wildcard);
  }

  /**
   * Implements DrupalCacheInterface::getMultiple().
   */
  function getMultiple(&$cids) {
    $cache = array();
    if ($last_write_time = $this->lastWriteTime()) {
      foreach ($this->inconsistentBackend->getMultiple($cids) as $item) {
        if ($item->created < $last_write_time) {
          $cids[] = $item->cid;
        }
        else {
          $cache[$item->cid] = $item;
        }
      }
    }

    if ($cids) {
      $items = $this->consistentBackend->getMultiple($cids);
      foreach ($items as $item) {
        $cache[$item->cid] = $item;
        $this->inconsistentBackend->set($item->cid, $item->data);
      }
    }

    return $cache;
  }

  /**
   * Get the last write time.
   */
  public function lastWriteTime() {
    if ($this->lastWriteTime === NULL) {
      $result = db_query('SELECT value FROM {schrodicache_variable} WHERE name = :name', array(':name' => self::LAST_WRITE_TIME_PREFIX . $this->bin))->fetchField();
      if ($result) {
        $this->lastWriteTime = unserialize($result);
      }
      else {
        $this->markAsOutdated();
      }
    }
    return $this->lastWriteTime;
  }

  protected function setLastWriteTime($time) {
    $this->lastWriteTime = $time;
  }

  public static function updateLastWriteTimes(array $last_write_times) {
    $transaction = db_transaction();
    try {
      db_query('DELETE FROM {schrodicache_variable} WHERE name IN (:names)', array(':names' => array_keys($last_write_times)));
      $insert = db_insert('schrodicache_variable');
      foreach ($last_write_times as $name => $value) {
        $insert->fields(array(
          'name' => $name,
          'value' => serialize($value),
        ));
      }
      $insert->execute();
    }
    catch (Exception $e) {
      $transaction->rollback();
    }
  }

  /**
   * Mark the inconsistent cache bin as outdated because of a write.
   */
  protected function markAsOutdated() {
    $value = round(microtime(TRUE), 3);
    $name = self::LAST_WRITE_TIME_PREFIX . $this->bin;
    if ($this->immediateLastWriteTimeUpdates) {
      static::updateLastWriteTimes(array($name => $value));
    }
    else {
      static::$writeTimes[$name] = $value;
    }

    $this->setLastWriteTime($value);
  }
}

